# -*- coding:utf-8 -*-
from django.db import connections
from django.db.utils import DEFAULT_DB_ALIAS

from django_mysql.exceptions import TimeoutError


class Lock(object):
    def __init__(self, name, acquire_timeout=10.0, using=None):
        self.acquire_timeout = acquire_timeout

        if using is None:
            self.db = DEFAULT_DB_ALIAS
        else:
            self.db = using

        # For multi-database servers, we prefix the name of the lock wth
        # the database, to protect against concurrent apps with the same locks
        self.name = self.make_name(self.db, name)

    @classmethod
    def make_name(cls, db, name):
        return '.'.join((
            connections[db].settings_dict['NAME'],
            name
        ))

    @classmethod
    def unmake_name(cls, db, name):
        # Cut off the 'dbname.' prefix
        db_name = connections[db].settings_dict['NAME']
        return name[len(db_name) + 1:]

    def get_cursor(self):
        return connections[self.db].cursor()

    def __enter__(self):
        with self.get_cursor() as cursor:
            cursor.execute(
                "SELECT GET_LOCK(%s, %s)",
                (self.name, self.acquire_timeout)
            )
            result = cursor.fetchone()[0]
            if result == 1:
                return self
            else:
                raise TimeoutError(
                    "Waited >{} seconds to gain lock".format(
                        self.acquire_timeout)
                )

    def __exit__(self, a, b, c):
        with self.get_cursor() as cursor:
            cursor.execute("SELECT RELEASE_LOCK(%s)", (self.name,))
            result = cursor.fetchone()[0]

            if result is None or result == 0:
                raise ValueError("Tried to release an unheld lock.")

    def is_held(self):
        return (self.holding_connection_id() is not None)

    def holding_connection_id(self):
        with self.get_cursor() as cursor:
            cursor.execute("SELECT IS_USED_LOCK(%s)", (self.name,))
            return cursor.fetchone()[0]

    @classmethod
    def held_with_prefix(cls, prefix, using=DEFAULT_DB_ALIAS):
        # Use the METADATA_LOCK_INFO table from the MariaDB plugin to show
        # which locks of a given prefix are held
        prefix = cls.make_name(using, prefix)

        with connections[using].cursor() as cursor:
            cursor.execute(
                """SELECT TABLE_SCHEMA, THREAD_ID
                   FROM INFORMATION_SCHEMA.METADATA_LOCK_INFO
                   WHERE TABLE_SCHEMA LIKE %s AND
                         LOCK_TYPE = 'User Lock'""",
                (prefix + '%',)
            )
            return {
                cls.unmake_name(using, row[0]): row[1]
                for row in cursor.fetchall()
            }



# -*- coding:utf-8 -*-
from django.test import TestCase

from django_mysql.exceptions import TimeoutError
from django_mysql.status import (
    GlobalStatus, SessionStatus, global_status, session_status
)


class GlobalStatusTests(TestCase):

    def test_get(self):
        running = global_status.get('Threads_running')
        self.assertTrue(isinstance(running, int))
        self.assertGreaterEqual(running, 1)

        cost = global_status.get('Last_query_cost')
        self.assertTrue(isinstance(cost, float))
        self.assertGreaterEqual(cost, 0.00)

        with self.assertRaises(ValueError) as cm:
            global_status.get('foo%')
        self.assertIn('wildcards', str(cm.exception))

        with self.assertRaises(KeyError):
            global_status.get('Does_not_exist')

    def test_get_many(self):
        myvars = global_status.get_many([])
        self.assertEqual(myvars, {})

        myvars = global_status.get_many(['Threads_running', 'Uptime'])
        self.assertTrue(isinstance(myvars, dict))
        self.assertIn('Threads_running', myvars)
        self.assertTrue(isinstance(myvars['Threads_running'], int))
        self.assertIn('Uptime', myvars)
        self.assertTrue(isinstance(myvars['Uptime'], int))

        with self.assertRaises(ValueError) as cm:
            global_status.get_many(['foo%'])
        self.assertIn('wildcards', str(cm.exception))

    def test_as_dict(self):
        status_dict = global_status.as_dict()

        self.assertIn('Aborted_clients', status_dict)  # Global-only variable

        self.assertTrue(isinstance(status_dict['Threads_running'], int))
        self.assertGreaterEqual(status_dict['Threads_running'], 1)

        self.assertTrue(isinstance(status_dict['Last_query_cost'], float))
        self.assertGreaterEqual(status_dict['Last_query_cost'], 0.0)

        self.assertTrue(isinstance(status_dict['Compression'], bool))

    def test_as_dict_prefix(self):
        status_dict = global_status.as_dict()

        status_dict_threads = global_status.as_dict('Threads_')
        self.assertLess(len(status_dict_threads), len(status_dict))
        for key in status_dict_threads:
            self.assertTrue(key.startswith('Threads_'))

        status_dict_foo = global_status.as_dict('Foo_Non_Existent')
        self.assertEqual(len(status_dict_foo), 0)

    def test_wait_until_load_low(self):
        # Assume tests are running on a non-busy server
        global_status.wait_until_load_low()
        global_status.wait_until_load_low({'Threads_running': 50,
                                           'Threads_connected': 100})

        with self.assertRaises(TimeoutError) as cm:
            global_status.wait_until_load_low(
                {'Threads_running': -1},  # obviously impossible
                timeout=0.001,
                sleep=0.0005
            )
        message = str(cm.exception)
        self.assertIn('Threads_running', message)
        self.assertIn('-1', message)

        with self.assertRaises(TimeoutError) as cm:
            global_status.wait_until_load_low(
                {'Threads_running': 1000000,
                 'Uptime': -1},  # obviously impossible
                timeout=0.001,
                sleep=0.0005
            )
        message = str(cm.exception)
        self.assertIn('Uptime', message)
        self.assertIn('-1', message)
        self.assertNotIn('Threads_running', message)
        self.assertNotIn('1000000', message)

    def test_other_databases(self):
        status = GlobalStatus(using='other')

        running = status.get('Threads_running')
        self.assertTrue(isinstance(running, int))
        self.assertGreaterEqual(running, 1)


class SessionStatusTests(TestCase):

    def test_get(self):
        bytes_received = session_status.get('Bytes_received')
        self.assertTrue(isinstance(bytes_received, int))
        self.assertGreaterEqual(bytes_received, 0)

        bytes_received_2 = session_status.get('Bytes_received')
        self.assertGreaterEqual(bytes_received_2, bytes_received)

        cost = session_status.get('Last_query_cost')
        self.assertTrue(isinstance(cost, float))
        self.assertGreaterEqual(cost, 0.00)

        with self.assertRaises(ValueError) as cm:
            session_status.get('foo%')
        self.assertIn('wildcards', str(cm.exception))

        with self.assertRaises(KeyError):
            session_status.get('Does_not_exist')

    def test_as_dict(self):
        status_dict = session_status.as_dict()

        self.assertIn('Compression', status_dict)

        self.assertTrue(isinstance(status_dict['Threads_running'], int))
        self.assertGreaterEqual(status_dict['Threads_running'], 1)

    def test_other_databases(self):
        status = SessionStatus(using='other')
        running = status.get('Threads_running')
        self.assertTrue(isinstance(running, int))
        self.assertGreaterEqual(running, 1)



# -*- coding:utf-8 -*-
import json
import re
from unittest import skip, skipIf

import ddt
import django
from django import forms
from django.core import exceptions, serializers
from django.core.management import call_command
from django.db import connection, models
from django.db.migrations.writer import MigrationWriter
from django.db.models import Q
from django.test import TestCase, override_settings

from django_mysql.forms import SimpleListField
from django_mysql.models import ListCharField, ListF
from django_mysql.test.utils import override_mysql_variables
from django_mysql_tests.models import (
    CharListDefaultModel, CharListModel, IntListModel
)


@ddt.ddt
class TestSaveLoad(TestCase):

    def test_char_easy(self):
        s = CharListModel.objects.create(field=["comfy", "big"])
        self.assertEqual(s.field, ["comfy", "big"])
        s = CharListModel.objects.get(id=s.id)
        self.assertEqual(s.field, ["comfy", "big"])

        s.field.append("round")
        s.save()
        self.assertEqual(s.field, ["comfy", "big", "round"])
        s = CharListModel.objects.get(id=s.id)
        self.assertEqual(s.field, ["comfy", "big", "round"])

    def test_char_string_direct(self):
        s = CharListModel.objects.create(field="big,bad")
        s = CharListModel.objects.get(id=s.id)
        self.assertEqual(s.field, ['big', 'bad'])

    def test_is_a_list_immediately(self):
        s = CharListModel()
        self.assertEqual(s.field, [])
        s.field.append("bold")
        s.field.append("brave")
        s.save()
        self.assertEqual(s.field, ["bold", "brave"])
        s = CharListModel.objects.get(id=s.id)
        self.assertEqual(s.field, ["bold", "brave"])

    def test_empty(self):
        s = CharListModel.objects.create()
        self.assertEqual(s.field, [])
        s = CharListModel.objects.get(id=s.id)
        self.assertEqual(s.field, [])

    def test_char_cant_create_lists_with_empty_string(self):
        with self.assertRaises(ValueError):
            CharListModel.objects.create(field=[""])

    def test_char_cant_create_sets_with_commas(self):
        with self.assertRaises(ValueError):
            CharListModel.objects.create(field=["co,mma", "contained"])

    def test_char_basic_lookup(self):
        mymodel = CharListModel.objects.create()
        empty = CharListModel.objects.filter(field="")

        self.assertEqual(empty.count(), 1)
        self.assertEqual(empty[0], mymodel)

        mymodel.delete()

        self.assertEqual(empty.count(), 0)

    @ddt.data('contains', 'icontains')
    def test_char_lookup(self, lookup):
        lname = 'field__' + lookup
        mymodel = CharListModel.objects.create(field=["mouldy", "rotten"])

        mouldy = CharListModel.objects.filter(**{lname: "mouldy"})
        self.assertEqual(mouldy.count(), 1)
        self.assertEqual(mouldy[0], mymodel)

        rotten = CharListModel.objects.filter(**{lname: "rotten"})
        self.assertEqual(rotten.count(), 1)
        self.assertEqual(rotten[0], mymodel)

        clean = CharListModel.objects.filter(**{lname: "clean"})
        self.assertEqual(clean.count(), 0)

        with self.assertRaises(ValueError):
            list(CharListModel.objects.filter(**{lname: ["a", "b"]}))

        both = CharListModel.objects.filter(
            Q(**{lname: "mouldy"}) & Q(**{lname: "rotten"})
        )
        self.assertEqual(both.count(), 1)
        self.assertEqual(both[0], mymodel)

        either = CharListModel.objects.filter(
            Q(**{lname: "mouldy"}) | Q(**{lname: "clean"})
        )
        self.assertEqual(either.count(), 1)

        not_clean = CharListModel.objects.exclude(**{lname: "clean"})
        self.assertEqual(not_clean.count(), 1)

        not_mouldy = CharListModel.objects.exclude(**{lname: "mouldy"})
        self.assertEqual(not_mouldy.count(), 0)

    def test_char_len_lookup_empty(self):
        mymodel = CharListModel.objects.create(field=[])

        empty = CharListModel.objects.filter(field__len=0)
        self.assertEqual(empty.count(), 1)
        self.assertEqual(empty[0], mymodel)

        one = CharListModel.objects.filter(field__len=1)
        self.assertEqual(one.count(), 0)

        one_or_more = CharListModel.objects.filter(field__len__gte=0)
        self.assertEqual(one_or_more.count(), 1)

    def test_char_len_lookup(self):
        mymodel = CharListModel.objects.create(field=["red", "expensive"])

        empty = CharListModel.objects.filter(field__len=0)
        self.assertEqual(empty.count(), 0)

        one_or_more = CharListModel.objects.filter(field__len__gte=1)
        self.assertEqual(one_or_more.count(), 1)
        self.assertEqual(one_or_more[0], mymodel)

        two = CharListModel.objects.filter(field__len=2)
        self.assertEqual(two.count(), 1)
        self.assertEqual(two[0], mymodel)

        three = CharListModel.objects.filter(field__len=3)
        self.assertEqual(three.count(), 0)

    def test_char_default(self):
        mymodel = CharListDefaultModel.objects.create()
        self.assertEqual(mymodel.field, ["a", "d"])

        mymodel = CharListDefaultModel.objects.get(id=mymodel.id)
        self.assertEqual(mymodel.field, ["a", "d"])

    def test_char_position_lookup(self):
        mymodel = CharListModel.objects.create(field=["red", "blue"])

        blue0 = CharListModel.objects.filter(field__0="blue")
        self.assertEqual(blue0.count(), 0)

        red0 = CharListModel.objects.filter(field__0="red")
        self.assertEqual(list(red0), [mymodel])

        red0_red1 = CharListModel.objects.filter(field__0="red",
                                                 field__1="red")
        self.assertEqual(red0_red1.count(), 0)

        red0_blue1 = CharListModel.objects.filter(field__0="red",
                                                  field__1="blue")
        self.assertEqual(list(red0_blue1), [mymodel])

        red0_or_blue0 = CharListModel.objects.filter(
            Q(field__0="red") | Q(field__0="blue")
        )
        self.assertEqual(list(red0_or_blue0), [mymodel])

    def test_char_position_lookup_repeat_fails(self):
        """
        FIND_IN_SET returns the *first* position so repeats are not dealt with
        """
        CharListModel.objects.create(field=["red", "red", "blue"])

        red1 = CharListModel.objects.filter(field__1="red")
        self.assertEqual(list(red1), [])  # should be 'red'

    def test_char_position_lookup_too_long(self):
        CharListModel.objects.create(field=["red", "blue"])

        red1 = CharListModel.objects.filter(field__2="blue")
        self.assertEqual(list(red1), [])

    def test_int_easy(self):
        mymodel = IntListModel.objects.create(field=[1, 2])
        self.assertEqual(mymodel.field, [1, 2])
        mymodel = IntListModel.objects.get(id=mymodel.id)
        self.assertEqual(mymodel.field, [1, 2])

    def test_int_contains_lookup(self):
        onetwo = IntListModel.objects.create(field=[1, 2])

        ones = IntListModel.objects.filter(field__contains=1)
        self.assertEqual(ones.count(), 1)
        self.assertEqual(ones[0], onetwo)

        twos = IntListModel.objects.filter(field__contains=2)
        self.assertEqual(twos.count(), 1)
        self.assertEqual(twos[0], onetwo)

        threes = IntListModel.objects.filter(field__contains=3)
        self.assertEqual(threes.count(), 0)

        with self.assertRaises(ValueError):
            list(IntListModel.objects.filter(field__contains=[1, 2]))

        ones_and_twos = IntListModel.objects.filter(
            Q(field__contains=1) & Q(field__contains=2)
        )
        self.assertEqual(ones_and_twos.count(), 1)
        self.assertEqual(ones_and_twos[0], onetwo)

        ones_and_threes = IntListModel.objects.filter(
            Q(field__contains=1) & Q(field__contains=3)
        )
        self.assertEqual(ones_and_threes.count(), 0)

        ones_or_threes = IntListModel.objects.filter(
            Q(field__contains=1) | Q(field__contains=3)
        )
        self.assertEqual(ones_or_threes.count(), 1)

        no_three = IntListModel.objects.exclude(field__contains=3)
        self.assertEqual(no_three.count(), 1)

        no_one = IntListModel.objects.exclude(field__contains=1)
        self.assertEqual(no_one.count(), 0)

    def test_int_position_lookup(self):
        onetwo = IntListModel.objects.create(field=[1, 2])

        one0 = IntListModel.objects.filter(field__0=1)
        self.assertEqual(list(one0), [onetwo])

        two0 = IntListModel.objects.filter(field__0=2)
        self.assertEqual(two0.count(), 0)

        one0two1 = IntListModel.objects.filter(field__0=1, field__1=2)
        self.assertEqual(list(one0two1), [onetwo])


@skipIf(django.VERSION <= (1, 8),
        "Requires Expressions from Django 1.8+")
class TestListF(TestCase):

    def test_append_to_none(self):
        CharListModel.objects.create(field=[])
        CharListModel.objects.update(field=ListF('field').append('first'))
        model = CharListModel.objects.get()
        self.assertEqual(model.field, ["first"])

    def test_append_to_one(self):
        CharListModel.objects.create(field=["big"])
        CharListModel.objects.update(field=ListF('field').append('bad'))
        model = CharListModel.objects.get()
        self.assertEqual(model.field, ["big", "bad"])

    def test_append_to_some(self):
        CharListModel.objects.create(field=["big", "blue"])
        CharListModel.objects.update(field=ListF('field').append('round'))
        model = CharListModel.objects.get()
        self.assertEqual(model.field, ["big", "blue", "round"])

    def test_append_to_multiple_objects(self):
        CharListModel.objects.create(field=["mouse"])
        CharListModel.objects.create(field=["keyboard"])
        CharListModel.objects.update(field=ListF('field').append("screen"))
        first, second = tuple(CharListModel.objects.all())
        self.assertEqual(first.field, ["mouse", "screen"])
        self.assertEqual(second.field, ["keyboard", "screen"])

    def test_append_exists(self):
        CharListModel.objects.create(field=["nice"])
        CharListModel.objects.update(field=ListF('field').append("nice"))
        model = CharListModel.objects.get()
        self.assertEqual(model.field, ["nice", "nice"])

    @override_mysql_variables(SQL_MODE="ANSI")
    def test_append_works_in_ansi_mode(self):
        CharListModel.objects.create()
        CharListModel.objects.update(field=ListF('field').append('big'))
        CharListModel.objects.update(field=ListF('field').append('bad'))
        model = CharListModel.objects.get()
        self.assertEqual(model.field, ["big", "bad"])

    def test_append_assignment(self):
        model = CharListModel.objects.create(field=["red"])
        model.field = ListF('field').append('blue')
        model.save()
        model = CharListModel.objects.get()
        self.assertEqual(model.field, ['red', 'blue'])

    def test_appendleft_to_none(self):
        CharListModel.objects.create(field=[])
        CharListModel.objects.update(field=ListF('field').appendleft('first'))
        model = CharListModel.objects.get()
        self.assertEqual(model.field, ["first"])

    def test_appendleft_to_one(self):
        CharListModel.objects.create(field=["big"])
        CharListModel.objects.update(field=ListF('field').appendleft('bad'))
        model = CharListModel.objects.get()
        self.assertEqual(model.field, ["bad", "big"])

    def test_appendleft_to_some(self):
        CharListModel.objects.create(field=["big", "blue"])
        CharListModel.objects.update(field=ListF('field').appendleft('round'))
        model = CharListModel.objects.get()
        self.assertEqual(model.field, ["round", "big", "blue"])

    def test_appendleft_to_multiple_objects(self):
        CharListModel.objects.create(field=["mouse"])
        CharListModel.objects.create(field=["keyboard"])
        CharListModel.objects.update(field=ListF('field').appendleft("screen"))
        first, second = tuple(CharListModel.objects.all())
        self.assertEqual(first.field, ["screen", "mouse"])
        self.assertEqual(second.field, ["screen", "keyboard"])

    def test_appendleft_exists(self):
        CharListModel.objects.create(field=["nice"])
        CharListModel.objects.update(field=ListF('field').appendleft("nice"))
        model = CharListModel.objects.get()
        self.assertEqual(model.field, ["nice", "nice"])

    @override_mysql_variables(SQL_MODE="ANSI")
    def test_appendleft_works_in_ansi_mode(self):
        CharListModel.objects.create()
        CharListModel.objects.update(field=ListF('field').appendleft('big'))
        CharListModel.objects.update(field=ListF('field').appendleft('bad'))
        model = CharListModel.objects.get()
        self.assertEqual(model.field, ["bad", "big"])

    def test_appendleft_assignment(self):
        model = CharListModel.objects.create(field=["red"])
        model.field = ListF('field').appendleft('blue')
        model.save()
        model = CharListModel.objects.get()
        self.assertEqual(model.field, ['blue', 'red'])

    def test_pop_none(self):
        CharListModel.objects.create(field=[])
        CharListModel.objects.update(field=ListF('field').pop())
        model = CharListModel.objects.get()
        self.assertEqual(model.field, [])

    def test_pop_one(self):
        CharListModel.objects.create(field=["red"])
        CharListModel.objects.update(field=ListF('field').pop())
        model = CharListModel.objects.get()
        self.assertEqual(model.field, [])

    def test_pop_two(self):
        CharListModel.objects.create(field=["red", "blue"])
        CharListModel.objects.update(field=ListF('field').pop())
        model = CharListModel.objects.get()
        self.assertEqual(model.field, ["red"])

    def test_pop_three(self):
        CharListModel.objects.create(field=["green", "yellow", "p"])
        CharListModel.objects.update(field=ListF('field').pop())
        model = CharListModel.objects.get()
        self.assertEqual(model.field, ["green", "yellow"])

    def test_popleft_none(self):
        CharListModel.objects.create(field=[])
        CharListModel.objects.update(field=ListF('field').popleft())
        model = CharListModel.objects.get()
        self.assertEqual(model.field, [])

    def test_popleft_one(self):
        CharListModel.objects.create(field=["red"])
        CharListModel.objects.update(field=ListF('field').popleft())
        model = CharListModel.objects.get()
        self.assertEqual(model.field, [])

    def test_popleft_two(self):
        CharListModel.objects.create(field=["red", "blue"])
        CharListModel.objects.update(field=ListF('field').popleft())
        model = CharListModel.objects.get()
        self.assertEqual(model.field, ["blue"])

    def test_popleft_three(self):
        CharListModel.objects.create(field=["green", "yellow", "p"])
        CharListModel.objects.update(field=ListF('field').popleft())
        model = CharListModel.objects.get()
        self.assertEqual(model.field, ["yellow", "p"])


class TestValidation(TestCase):

    def test_max_length(self):
        field = ListCharField(
            models.CharField(max_length=32),
            size=3,
            max_length=32
        )

        field.clean({'a', 'b', 'c'}, None)

        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean({'a', 'b', 'c', 'd'}, None)
        self.assertEqual(
            cm.exception.messages[0],
            'List contains 4 items, it should contain no more than 3.'
        )


class TestCheck(TestCase):

    def test_field_checks(self):
        field = ListCharField(models.CharField(), max_length=32)
        field.set_attributes_from_name('field')
        errors = field.check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, 'django_mysql.E004')
        self.assertIn('Base field for list has errors', errors[0].msg)
        self.assertIn('max_length', errors[0].msg)

    def test_invalid_base_fields(self):
        field = ListCharField(
            models.ForeignKey('django_mysql_tests.Author'),
            max_length=32
        )
        field.set_attributes_from_name('field')
        errors = field.check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, 'django_mysql.E005')
        self.assertIn('Base field for list must be', errors[0].msg)

    def test_max_length_including_base(self):
        field = ListCharField(
            models.CharField(max_length=32),
            size=2, max_length=32)
        field.set_attributes_from_name('field')
        errors = field.check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, 'django_mysql.E006')
        self.assertIn('Field can overrun', errors[0].msg)


class TestMigrations(TestCase):

    def test_deconstruct(self):
        field = ListCharField(models.IntegerField(), max_length=32)
        name, path, args, kwargs = field.deconstruct()
        new = ListCharField(*args, **kwargs)
        self.assertEqual(type(new.base_field), type(field.base_field))

    def test_deconstruct_with_size(self):
        field = ListCharField(models.IntegerField(), size=3, max_length=32)
        name, path, args, kwargs = field.deconstruct()
        new = ListCharField(*args, **kwargs)
        self.assertEqual(new.size, field.size)

    def test_deconstruct_args(self):
        field = ListCharField(models.CharField(max_length=5), max_length=32)
        name, path, args, kwargs = field.deconstruct()
        new = ListCharField(*args, **kwargs)
        self.assertEqual(
            new.base_field.max_length,
            field.base_field.max_length
        )

    def test_makemigrations(self):
        field = ListCharField(models.CharField(max_length=5), max_length=32)
        statement, imports = MigrationWriter.serialize(field)

        # The order of the output max_length/size statements varies by
        # python version, hence a little regexp to match them
        self.assertRegexpMatches(
            statement,
            re.compile(
                r"""^django_mysql\.models\.ListCharField\(
                    models\.CharField\(max_length=5\),\ # space here
                    (
                        max_length=32,\ size=None|
                        size=None,\ max_length=32
                    )
                    \)$
                """,
                re.VERBOSE
            )
        )

    def test_makemigrations_with_size(self):
        field = ListCharField(
            models.CharField(max_length=5),
            max_length=32,
            size=5
        )
        statement, imports = MigrationWriter.serialize(field)

        # The order of the output max_length/size statements varies by
        # python version, hence a little regexp to match them
        self.assertRegexpMatches(
            statement,
            re.compile(
                r"""^django_mysql\.models\.ListCharField\(
                    models\.CharField\(max_length=5\),\ # space here
                    (
                        max_length=32,\ size=5|
                        size=5,\ max_length=32
                    )
                    \)$
                """,
                re.VERBOSE
            )
        )

    @skip("Unforunately this and the SetCharField equivalent interfere with "
          "each other - the migrations don't seem to roll back smoothly.")
    @override_settings(MIGRATION_MODULES={
        "django_mysql_tests": "django_mysql_tests.list_default_migrations",
    })
    def test_adding_field_with_default(self):
        table_name = 'django_mysql_tests_intlistdefaultmodel'
        with connection.cursor() as cursor:
            self.assertNotIn(
                table_name,
                connection.introspection.table_names(cursor)
            )

        call_command('migrate', 'django_mysql_tests', verbosity=0)
        with connection.cursor() as cursor:
            self.assertIn(
                table_name,
                connection.introspection.table_names(cursor)
            )

        call_command('migrate', 'django_mysql_tests', 'zero', verbosity=0)
        with connection.cursor() as cursor:
            self.assertNotIn(
                table_name,
                connection.introspection.table_names(cursor)
            )


class TestSerialization(TestCase):

    def test_dumping(self):
        instance = CharListModel(field=["big", "comfy"])
        data = json.loads(serializers.serialize('json', [instance]))[0]
        field = data['fields']['field']
        self.assertEqual(sorted(field.split(',')), ["big", "comfy"])

    def test_loading(self):
        test_data = '''
            [{"fields": {"field": "big,leather,comfy"},
             "model": "django_mysql_tests.CharListModel", "pk": null}]
        '''
        objs = list(serializers.deserialize('json', test_data))
        instance = objs[0].object
        self.assertEqual(instance.field, ["big", "leather", "comfy"])


class TestDescription(TestCase):

    def test_char(self):
        field = ListCharField(models.CharField(max_length=5), max_length=32)
        self.assertEqual(
            field.description,
            "List of String (up to %(max_length)s)"
        )

    def test_int(self):
        field = ListCharField(models.IntegerField(), max_length=32)
        self.assertEqual(field.description, "List of Integer")


class TestFormField(TestCase):

    def test_model_field_formfield(self):
        model_field = ListCharField(models.CharField(max_length=27))
        form_field = model_field.formfield()
        self.assertIsInstance(form_field, SimpleListField)
        self.assertIsInstance(form_field.base_field, forms.CharField)
        self.assertEqual(form_field.base_field.max_length, 27)

    def test_model_field_formfield_size(self):
        model_field = ListCharField(models.IntegerField(), size=4)
        form_field = model_field.formfield()
        self.assertIsInstance(form_field, SimpleListField)
        self.assertEqual(form_field.max_length, 4)



# -*- coding:utf-8 -*-
from django import forms
from django.core import exceptions
from django.test import TestCase

from django_mysql.forms import SimpleListField, SimpleSetField


class TestSimpleListField(TestCase):

    def test_valid(self):
        field = SimpleListField(forms.CharField())
        value = field.clean('a,b,c')
        self.assertEqual(value, ['a', 'b', 'c'])

    def test_to_python_no_leading_commas(self):
        field = SimpleListField(forms.IntegerField())
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean(',1')
        self.assertEqual(
            cm.exception.messages[0],
            'No leading, trailing, or double commas.'
        )

    def test_to_python_no_trailing_commas(self):
        field = SimpleListField(forms.IntegerField())
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('1,')
        self.assertEqual(
            cm.exception.messages[0],
            'No leading, trailing, or double commas.'
        )

    def test_to_python_no_double_commas(self):
        field = SimpleListField(forms.IntegerField())
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('1,,2')
        self.assertEqual(
            cm.exception.messages[0],
            'No leading, trailing, or double commas.'
        )

    def test_to_python_base_field_does_not_validate(self):
        field = SimpleListField(forms.IntegerField())
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('a,b,9')
        self.assertEqual(
            cm.exception.messages[0],
            'Item 1 in the list did not validate: Enter a whole number.'
        )

    def test_validate_fail(self):
        field = SimpleListField(
            forms.ChoiceField(choices=(('a', 'The letter A'),
                                       ('b', 'The letter B')))
        )
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('a,c')
        self.assertEqual(
            cm.exception.messages[0],
            'Item 2 in the list did not validate: '
            'Select a valid choice. c is not one of the available choices.'
        )

    def test_validators_fail(self):
        field = SimpleListField(forms.RegexField('[a-e]{2}'))
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('a,bc,de')
        self.assertEqual(
            cm.exception.messages[0],
            'Item 1 in the list did not validate: Enter a valid value.'
        )

    def test_validators_fail_base_max_length(self):
        field = SimpleListField(forms.CharField(max_length=5))
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('longer,yes')
        self.assertEqual(
            cm.exception.messages[0],
            'Item 1 in the list did not validate: '
            'Ensure this value has at most 5 characters (it has 6).'
        )

    def test_validators_fail_base_min_max_length(self):
        # there's just no satisfying some people...
        field = SimpleListField(forms.CharField(min_length=10, max_length=8))
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('undefined')
        self.assertEqual(
            cm.exception.messages[0],
            'Item 1 in the list did not validate: '
            'Ensure this value has at least 10 characters (it has 9).'
        )
        self.assertEqual(
            cm.exception.messages[1],
            'Item 1 in the list did not validate: '
            'Ensure this value has at most 8 characters (it has 9).'
        )

    def test_prepare_value(self):
        field = SimpleListField(forms.CharField())
        value = field.prepare_value(['a', 'b', 'c'])
        self.assertEqual(
            value.split(','),
            ['a', 'b', 'c']
        )

        self.assertEqual(field.prepare_value('1,a'), '1,a')

    def test_max_length(self):
        field = SimpleListField(forms.CharField(), max_length=2)
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('a,b,c')
        self.assertEqual(
            cm.exception.messages[0],
            'List contains 3 items, it should contain no more than 2.'
        )

    def test_min_length(self):
        field = SimpleListField(forms.CharField(), min_length=4)
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('a,b,c')
        self.assertEqual(
            cm.exception.messages[0],
            'List contains 3 items, it should contain no fewer than 4.'
        )

    def test_required(self):
        field = SimpleListField(forms.CharField(), required=True)
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('')
        self.assertEqual(cm.exception.messages[0], 'This field is required.')


class TestSimpleSetField(TestCase):

    def test_valid(self):
        field = SimpleSetField(forms.CharField())
        value = field.clean('a,b,c')
        self.assertEqual(value, {'a', 'b', 'c'})

    def test_to_python_no_leading_commas(self):
        field = SimpleSetField(forms.IntegerField())
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean(',1')
        self.assertEqual(
            cm.exception.messages[0],
            'No leading, trailing, or double commas.'
        )

    def test_to_python_no_trailing_commas(self):
        field = SimpleSetField(forms.IntegerField())
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('1,')
        self.assertEqual(
            cm.exception.messages[0],
            'No leading, trailing, or double commas.'
        )

    def test_to_python_no_double_commas(self):
        field = SimpleSetField(forms.IntegerField())
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('1,,2')
        self.assertEqual(
            cm.exception.messages[0],
            'No leading, trailing, or double commas.'
        )

    def test_to_python_base_field_does_not_validate(self):
        field = SimpleSetField(forms.IntegerField())
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('a,b,9')
        self.assertEqual(
            cm.exception.messages[0],
            'Item 1 in the set did not validate: Enter a whole number.'
        )

    def test_to_python_duplicates_not_allowed(self):
        field = SimpleSetField(forms.IntegerField())
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('1,1')
        self.assertEqual(
            cm.exception.messages[0],
            "Duplicates are not supported. '1' appears twice or more."
        )

    def test_to_python_two_duplicates_not_allowed(self):
        field = SimpleSetField(forms.IntegerField())
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('1,2,1,2')
        self.assertEqual(
            cm.exception.messages[0],
            "Duplicates are not supported. '1' appears twice or more."
        )
        self.assertEqual(
            cm.exception.messages[1],
            "Duplicates are not supported. '2' appears twice or more."
        )

    def test_validate_fail(self):
        field = SimpleSetField(
            forms.ChoiceField(choices=(('a', 'The letter A'),
                                       ('b', 'The letter B')))
        )
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('a,c')
        self.assertEqual(
            cm.exception.messages[0],
            'Item "c" in the set did not validate: '
            'Select a valid choice. c is not one of the available choices.'
        )

    def test_validators_fail(self):
        field = SimpleSetField(forms.RegexField('[a-e]{2}'))
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('a,bc,de')
        self.assertEqual(
            cm.exception.messages[0],
            'Item "a" in the set did not validate: Enter a valid value.'
        )

    def test_validators_fail_base_max_length(self):
        field = SimpleSetField(forms.CharField(max_length=5))
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('longer,yes')
        self.assertEqual(
            cm.exception.messages[0],
            'Item "longer" in the set did not validate: '
            'Ensure this value has at most 5 characters (it has 6).'
        )

    def test_validators_fail_base_min_max_length(self):
        # there's just no satisfying some people...
        field = SimpleSetField(forms.CharField(min_length=10, max_length=8))
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('undefined')
        self.assertEqual(
            cm.exception.messages[0],
            'Item "undefined" in the set did not validate: '
            'Ensure this value has at least 10 characters (it has 9).'
        )
        self.assertEqual(
            cm.exception.messages[1],
            'Item "undefined" in the set did not validate: '
            'Ensure this value has at most 8 characters (it has 9).'
        )

    def test_prepare_value(self):
        field = SimpleSetField(forms.CharField())
        value = field.prepare_value({'a', 'b', 'c'})
        self.assertEqual(
            list(sorted(value.split(','))),
            ['a', 'b', 'c']
        )

        self.assertEqual(field.prepare_value('1,a'), '1,a')

    def test_max_length(self):
        field = SimpleSetField(forms.CharField(), max_length=2)
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('a,b,c')
        self.assertEqual(
            cm.exception.messages[0],
            'Set contains 3 items, it should contain no more than 2.'
        )

    def test_min_length(self):
        field = SimpleSetField(forms.CharField(), min_length=4)
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('a,b,c')
        self.assertEqual(
            cm.exception.messages[0],
            'Set contains 3 items, it should contain no fewer than 4.'
        )

    def test_required(self):
        field = SimpleSetField(forms.CharField(), required=True)
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('')
        self.assertEqual(cm.exception.messages[0], 'This field is required.')



# -*- coding:utf-8 -*-
import os

import django

BASE_DIR = os.path.dirname(os.path.dirname(__file__))
DEBUG = True
TEMPLATE_DEBUG = DEBUG

SECRET_KEY = 'THISuISdNOT9A$SECRET9x&ji!vceayg+wwt472!bgs$0!i3k4'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.mysql',
        'NAME': 'django_mysql',
        'USER': '',
        'PASSWORD': '',
        'HOST': '',
        'PORT': '',
        'OPTIONS': {'charset': 'utf8mb4'},
        'TEST': {
            'COLLATION': "utf8mb4_general_ci",
            'CHARSET': "utf8mb4"
        }
    },
    'other': {
        'ENGINE': 'django.db.backends.mysql',
        'NAME': 'django_mysql2',
        'USER': '',
        'PASSWORD': '',
        'HOST': '',
        'PORT': '',
        'OPTIONS': {'charset': 'utf8mb4'},
        'TEST': {
            'COLLATION': "utf8mb4_general_ci",
            'CHARSET': "utf8mb4"
        }
    },
}

ALLOWED_HOSTS = []

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django_mysql',
    'django_mysql_tests'
)

if django.VERSION[:2] >= (1, 7):
    INSTALLED_APPS = (
        'django.contrib.admin.apps.AdminConfig',
    ) + INSTALLED_APPS
else:
    INSTALLED_APPS = (
        'django.contrib.admin',
    ) + INSTALLED_APPS

MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'urls'
LANGUAGE_CODE = 'en-us'
TIME_ZONE = 'UTC'
USE_I18N = True
USE_L10N = True
STATIC_URL = '/static/'

try:
    from local_settings import *  # noqa
except ImportError:
    pass

TEST_RUNNER = 'django_mysql_tests.runner.MySQLTestRunner'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [],
        },
    },
]



# -*- coding:utf-8 -*-
import json
import re
from unittest import skipIf

import ddt
import django
from django import forms
from django.core import exceptions, serializers
from django.core.management import call_command
from django.db import connection, models
from django.db.migrations.writer import MigrationWriter
from django.db.models import Q
from django.test import TestCase, override_settings

from django_mysql.forms import SimpleSetField
from django_mysql.models import SetCharField, SetF
from django_mysql.test.utils import override_mysql_variables
from django_mysql_tests.models import (
    CharSetDefaultModel, CharSetModel, IntSetModel
)


@ddt.ddt
class TestSaveLoad(TestCase):

    def test_char_easy(self):
        s = CharSetModel.objects.create(field={"big", "comfy"})
        self.assertEqual(s.field, {"comfy", "big"})
        s = CharSetModel.objects.get(id=s.id)
        self.assertEqual(s.field, {"comfy", "big"})

        s.field.add("round")
        s.save()
        self.assertEqual(s.field, {"comfy", "big", "round"})
        s = CharSetModel.objects.get(id=s.id)
        self.assertEqual(s.field, {"comfy", "big", "round"})

    def test_char_string_direct(self):
        s = CharSetModel.objects.create(field="big,bad")
        s = CharSetModel.objects.get(id=s.id)
        self.assertEqual(s.field, {'big', 'bad'})

    def test_is_a_set_immediately(self):
        s = CharSetModel()
        self.assertEqual(s.field, set())
        s.field.add("bold")
        s.field.add("brave")
        s.save()
        self.assertEqual(s.field, {"bold", "brave"})
        s = CharSetModel.objects.get(id=s.id)
        self.assertEqual(s.field, {"bold", "brave"})

    def test_empty(self):
        s = CharSetModel.objects.create()
        self.assertEqual(s.field, set())
        s = CharSetModel.objects.get(id=s.id)
        self.assertEqual(s.field, set())

    def test_char_cant_create_sets_with_empty_string(self):
        with self.assertRaises(ValueError):
            CharSetModel.objects.create(field={""})

    def test_char_cant_create_sets_with_commas(self):
        with self.assertRaises(ValueError):
            CharSetModel.objects.create(field={"co,mma", "contained"})

    def test_char_basic_lookup(self):
        mymodel = CharSetModel.objects.create()
        empty = CharSetModel.objects.filter(field="")

        self.assertEqual(empty.count(), 1)
        self.assertEqual(empty[0], mymodel)

        mymodel.delete()

        self.assertEqual(empty.count(), 0)

    @ddt.data('contains', 'icontains')
    def test_char_lookup(self, lookup):
        lname = 'field__' + lookup
        mymodel = CharSetModel.objects.create(field={"mouldy", "rotten"})

        mouldy = CharSetModel.objects.filter(**{lname: "mouldy"})
        self.assertEqual(mouldy.count(), 1)
        self.assertEqual(mouldy[0], mymodel)

        rotten = CharSetModel.objects.filter(**{lname: "rotten"})
        self.assertEqual(rotten.count(), 1)
        self.assertEqual(rotten[0], mymodel)

        clean = CharSetModel.objects.filter(**{lname: "clean"})
        self.assertEqual(clean.count(), 0)

        with self.assertRaises(ValueError):
            list(CharSetModel.objects.filter(**{lname: {"a", "b"}}))

        both = CharSetModel.objects.filter(
            Q(**{lname: "mouldy"}) & Q(**{lname: "rotten"})
        )
        self.assertEqual(both.count(), 1)
        self.assertEqual(both[0], mymodel)

        either = CharSetModel.objects.filter(
            Q(**{lname: "mouldy"}) | Q(**{lname: "clean"})
        )
        self.assertEqual(either.count(), 1)

        not_clean = CharSetModel.objects.exclude(**{lname: "clean"})
        self.assertEqual(not_clean.count(), 1)

        not_mouldy = CharSetModel.objects.exclude(**{lname: "mouldy"})
        self.assertEqual(not_mouldy.count(), 0)

    def test_char_len_lookup_empty(self):
        mymodel = CharSetModel.objects.create(field=set())

        empty = CharSetModel.objects.filter(field__len=0)
        self.assertEqual(empty.count(), 1)
        self.assertEqual(empty[0], mymodel)

        one = CharSetModel.objects.filter(field__len=1)
        self.assertEqual(one.count(), 0)

        one_or_more = CharSetModel.objects.filter(field__len__gte=0)
        self.assertEqual(one_or_more.count(), 1)

    def test_char_len_lookup(self):
        mymodel = CharSetModel.objects.create(field={"red", "expensive"})

        empty = CharSetModel.objects.filter(field__len=0)
        self.assertEqual(empty.count(), 0)

        one_or_more = CharSetModel.objects.filter(field__len__gte=1)
        self.assertEqual(one_or_more.count(), 1)
        self.assertEqual(one_or_more[0], mymodel)

        two = CharSetModel.objects.filter(field__len=2)
        self.assertEqual(two.count(), 1)
        self.assertEqual(two[0], mymodel)

        three = CharSetModel.objects.filter(field__len=3)
        self.assertEqual(three.count(), 0)

    def test_char_default(self):
        mymodel = CharSetDefaultModel.objects.create()
        self.assertEqual(mymodel.field, {"a", "d"})

        mymodel = CharSetDefaultModel.objects.get(id=mymodel.id)
        self.assertEqual(mymodel.field, {"a", "d"})

    def test_int_easy(self):
        mymodel = IntSetModel.objects.create(field={1, 2})
        self.assertEqual(mymodel.field, {1, 2})
        mymodel = IntSetModel.objects.get(id=mymodel.id)
        self.assertEqual(mymodel.field, {1, 2})

    def test_int_contains_lookup(self):
        onetwo = IntSetModel.objects.create(field={1, 2})

        ones = IntSetModel.objects.filter(field__contains=1)
        self.assertEqual(ones.count(), 1)
        self.assertEqual(ones[0], onetwo)

        twos = IntSetModel.objects.filter(field__contains=2)
        self.assertEqual(twos.count(), 1)
        self.assertEqual(twos[0], onetwo)

        threes = IntSetModel.objects.filter(field__contains=3)
        self.assertEqual(threes.count(), 0)

        with self.assertRaises(ValueError):
            list(IntSetModel.objects.filter(field__contains={1, 2}))

        ones_and_twos = IntSetModel.objects.filter(
            Q(field__contains=1) & Q(field__contains=2)
        )
        self.assertEqual(ones_and_twos.count(), 1)
        self.assertEqual(ones_and_twos[0], onetwo)

        ones_and_threes = IntSetModel.objects.filter(
            Q(field__contains=1) & Q(field__contains=3)
        )
        self.assertEqual(ones_and_threes.count(), 0)

        ones_or_threes = IntSetModel.objects.filter(
            Q(field__contains=1) | Q(field__contains=3)
        )
        self.assertEqual(ones_or_threes.count(), 1)

        no_three = IntSetModel.objects.exclude(field__contains=3)
        self.assertEqual(no_three.count(), 1)

        no_one = IntSetModel.objects.exclude(field__contains=1)
        self.assertEqual(no_one.count(), 0)


@skipIf(django.VERSION <= (1, 8),
        "Requires Expressions from Django 1.8+")
class TestSetF(TestCase):

    def test_add_to_none(self):
        CharSetModel.objects.create(field=set())
        CharSetModel.objects.update(field=SetF('field').add('first'))
        model = CharSetModel.objects.get()
        self.assertEqual(model.field, {"first"})

    def test_add_to_one(self):
        CharSetModel.objects.create(field={"big"})
        CharSetModel.objects.update(field=SetF('field').add('bad'))
        model = CharSetModel.objects.get()
        self.assertEqual(model.field, {"big", "bad"})

    def test_add_to_some(self):
        CharSetModel.objects.create(field={"big", "blue"})
        CharSetModel.objects.update(field=SetF('field').add('round'))
        model = CharSetModel.objects.get()
        self.assertEqual(model.field, {"big", "blue", "round"})

    def test_add_to_multiple_objects(self):
        CharSetModel.objects.create(field={"mouse"})
        CharSetModel.objects.create(field={"keyboard"})
        CharSetModel.objects.update(field=SetF('field').add("screen"))
        first, second = tuple(CharSetModel.objects.all())
        self.assertEqual(first.field, {"mouse", "screen"})
        self.assertEqual(second.field, {"keyboard", "screen"})

    def test_add_exists(self):
        CharSetModel.objects.create(field={"nice"})
        CharSetModel.objects.update(field=SetF('field').add("nice"))
        model = CharSetModel.objects.get()
        self.assertEqual(model.field, {"nice"})

    @override_mysql_variables(SQL_MODE="ANSI")
    def test_add_works_in_ansi_mode(self):
        CharSetModel.objects.create()
        CharSetModel.objects.update(field=SetF('field').add('big'))
        CharSetModel.objects.update(field=SetF('field').add('bad'))
        model = CharSetModel.objects.get()
        self.assertEqual(model.field, {"big", "bad"})

    def test_add_assignment(self):
        model = CharSetModel.objects.create(field={"red"})
        model.field = SetF('field').add('blue')
        model.save()
        model = CharSetModel.objects.get()
        self.assertEqual(model.field, {'red', 'blue'})

    def test_remove_one(self):
        CharSetModel.objects.create(field={"dopey", "knifey"})
        CharSetModel.objects.update(field=SetF('field').remove('knifey'))
        model = CharSetModel.objects.get()
        self.assertEqual(model.field, {"dopey"})

    def test_remove_only_one(self):
        CharSetModel.objects.create(field={"pants"})
        CharSetModel.objects.update(field=SetF('field').remove('pants'))
        model = CharSetModel.objects.get()
        self.assertEqual(model.field, set())

    def test_remove_from_none(self):
        CharSetModel.objects.create(field=set())
        CharSetModel.objects.update(field=SetF("field").remove("jam"))
        model = CharSetModel.objects.get()
        self.assertEqual(model.field, set())

    def test_remove_first(self):
        CharSetModel.objects.create()
        CharSetModel.objects.update(field="a,b,c")
        CharSetModel.objects.update(field=SetF('field').remove('a'))
        model = CharSetModel.objects.get()
        self.assertEqual(model.field, {"b", "c"})

    def test_remove_middle(self):
        CharSetModel.objects.create()
        CharSetModel.objects.update(field="a,b,c")
        CharSetModel.objects.update(field=SetF('field').remove('b'))
        model = CharSetModel.objects.get()
        self.assertEqual(model.field, {"a", "c"})

    def test_remove_last(self):
        CharSetModel.objects.create()
        CharSetModel.objects.update(field="a,b,c")
        CharSetModel.objects.update(field=SetF('field').remove('c'))
        model = CharSetModel.objects.get()
        self.assertEqual(model.field, {"a", "b"})

    def test_remove_not_exists(self):
        CharSetModel.objects.create(field={"nice"})
        CharSetModel.objects.update(field=SetF("field").remove("naughty"))
        model = CharSetModel.objects.get()
        self.assertEqual(model.field, {"nice"})

    def test_remove_from_multiple_objects(self):
        CharSetModel.objects.create(field={"mouse", "chair"})
        CharSetModel.objects.create(field={"keyboard", "chair"})
        CharSetModel.objects.update(field=SetF('field').remove("chair"))
        first, second = tuple(CharSetModel.objects.all())
        self.assertEqual(first.field, {"mouse"})
        self.assertEqual(second.field, {"keyboard"})

    @override_mysql_variables(SQL_MODE="ANSI")
    def test_remove_works_in_ansi_mode(self):
        CharSetModel.objects.create(field={"bold"})
        CharSetModel.objects.update(field=SetF('field').remove('big'))
        CharSetModel.objects.update(field=SetF('field').remove('bold'))
        CharSetModel.objects.update(field=SetF('field').remove('bad'))
        model = CharSetModel.objects.get()
        self.assertEqual(model.field, set())

    def test_remove_assignment(self):
        model = IntSetModel.objects.create(field={24, 89})
        model.field = SetF('field').remove(89)
        model.save()
        model = IntSetModel.objects.get()
        self.assertEqual(model.field, {24})

    def test_works_with_two_fields(self):
        CharSetModel.objects.create(field={"snickers", "lion"},
                                    field2={"apple", "orange"})

        # Concurrent add
        CharSetModel.objects.update(field=SetF('field').add("mars"),
                                    field2=SetF('field2').add("banana"))
        model = CharSetModel.objects.get()
        self.assertEqual(model.field, {"snickers", "lion", "mars"})
        self.assertEqual(model.field2, {"apple", "orange", "banana"})

        # Concurrent add and remove
        CharSetModel.objects.update(field=SetF('field').add("reeses"),
                                    field2=SetF('field2').remove("banana"))
        model = CharSetModel.objects.get()
        self.assertEqual(model.field, {"snickers", "lion", "mars", "reeses"})
        self.assertEqual(model.field2, {"apple", "orange"})

        # Swap
        CharSetModel.objects.update(field=SetF('field').remove("lion"),
                                    field2=SetF('field2').remove("apple"))
        model = CharSetModel.objects.get()
        self.assertEqual(model.field, {"snickers", "mars", "reeses"})
        self.assertEqual(model.field2, {"orange"})


@skipIf(django.VERSION >= (1, 8),
        "Requires old Django version without Expressions")
class TestSetFFails(TestCase):

    def test_cannot_instantiate(self):
        with self.assertRaises(ValueError):
            SetF('field').add("something")


class TestValidation(TestCase):

    def test_max_length(self):
        field = SetCharField(
            models.CharField(max_length=32),
            size=3,
            max_length=32
        )

        field.clean({'a', 'b', 'c'}, None)

        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean({'a', 'b', 'c', 'd'}, None)
        self.assertEqual(
            cm.exception.messages[0],
            'Set contains 4 items, it should contain no more than 3.'
        )


class TestCheck(TestCase):

    def test_field_checks(self):
        field = SetCharField(models.CharField(), max_length=32)
        field.set_attributes_from_name('field')
        errors = field.check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, 'django_mysql.E001')
        self.assertIn('Base field for set has errors', errors[0].msg)
        self.assertIn('max_length', errors[0].msg)

    def test_invalid_base_fields(self):
        field = SetCharField(
            models.ForeignKey('django_mysql_tests.Author'),
            max_length=32
        )
        field.set_attributes_from_name('field')
        errors = field.check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, 'django_mysql.E002')
        self.assertIn('Base field for set must be', errors[0].msg)

    def test_max_length_including_base(self):
        field = SetCharField(
            models.CharField(max_length=32),
            size=2, max_length=32)
        field.set_attributes_from_name('field')
        errors = field.check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, 'django_mysql.E003')
        self.assertIn('Field can overrun', errors[0].msg)


class TestMigrations(TestCase):

    def test_deconstruct(self):
        field = SetCharField(models.IntegerField(), max_length=32)
        name, path, args, kwargs = field.deconstruct()
        new = SetCharField(*args, **kwargs)
        self.assertEqual(type(new.base_field), type(field.base_field))

    def test_deconstruct_with_size(self):
        field = SetCharField(models.IntegerField(), size=3, max_length=32)
        name, path, args, kwargs = field.deconstruct()
        new = SetCharField(*args, **kwargs)
        self.assertEqual(new.size, field.size)

    def test_deconstruct_args(self):
        field = SetCharField(models.CharField(max_length=5), max_length=32)
        name, path, args, kwargs = field.deconstruct()
        new = SetCharField(*args, **kwargs)
        self.assertEqual(
            new.base_field.max_length,
            field.base_field.max_length
        )

    def test_makemigrations(self):
        field = SetCharField(models.CharField(max_length=5), max_length=32)
        statement, imports = MigrationWriter.serialize(field)

        # The order of the output max_length/size statements varies by
        # python version, hence a little regexp to match them
        self.assertRegexpMatches(
            statement,
            re.compile(
                r"""^django_mysql\.models\.SetCharField\(
                    models\.CharField\(max_length=5\),\ # space here
                    (
                        max_length=32,\ size=None|
                        size=None,\ max_length=32
                    )
                    \)$
                """,
                re.VERBOSE
            )
        )

    def test_makemigrations_with_size(self):
        field = SetCharField(
            models.CharField(max_length=5),
            max_length=32,
            size=5
        )
        statement, imports = MigrationWriter.serialize(field)

        # The order of the output max_length/size statements varies by
        # python version, hence a little regexp to match them
        self.assertRegexpMatches(
            statement,
            re.compile(
                r"""^django_mysql\.models\.SetCharField\(
                    models\.CharField\(max_length=5\),\ # space here
                    (
                        max_length=32,\ size=5|
                        size=5,\ max_length=32
                    )
                    \)$
                """,
                re.VERBOSE
            )
        )

    @override_settings(MIGRATION_MODULES={
        "django_mysql_tests": "django_mysql_tests.set_default_migrations",
    })
    def test_adding_field_with_default(self):
        table_name = 'django_mysql_tests_intsetdefaultmodel'
        with connection.cursor() as cursor:
            self.assertNotIn(
                table_name,
                connection.introspection.table_names(cursor)
            )

        call_command('migrate', 'django_mysql_tests', verbosity=0)
        with connection.cursor() as cursor:
            self.assertIn(
                table_name,
                connection.introspection.table_names(cursor)
            )

        call_command('migrate', 'django_mysql_tests', 'zero', verbosity=0)
        with connection.cursor() as cursor:
            self.assertNotIn(
                table_name,
                connection.introspection.table_names(cursor)
            )


class TestSerialization(TestCase):

    def test_dumping(self):
        instance = CharSetModel(field={"big", "comfy"})
        data = json.loads(serializers.serialize('json', [instance]))[0]
        field = data['fields']['field']
        self.assertEqual(sorted(field.split(',')), ["big", "comfy"])

    def test_loading(self):
        test_data = '''
            [{"fields": {"field": "big,leather,comfy"},
             "model": "django_mysql_tests.CharSetModel", "pk": null}]
        '''
        objs = list(serializers.deserialize('json', test_data))
        instance = objs[0].object
        self.assertEqual(instance.field, {"big", "leather", "comfy"})


class TestDescription(TestCase):

    def test_char(self):
        field = SetCharField(models.CharField(max_length=5), max_length=32)
        self.assertEqual(
            field.description,
            "Set of String (up to %(max_length)s)"
        )

    def test_int(self):
        field = SetCharField(models.IntegerField(), max_length=32)
        self.assertEqual(field.description, "Set of Integer")


class TestFormField(TestCase):

    def test_model_field_formfield(self):
        model_field = SetCharField(models.CharField(max_length=27))
        form_field = model_field.formfield()
        self.assertIsInstance(form_field, SimpleSetField)
        self.assertIsInstance(form_field.base_field, forms.CharField)
        self.assertEqual(form_field.base_field.max_length, 27)

    def test_model_field_formfield_size(self):
        model_field = SetCharField(models.IntegerField(), size=4)
        form_field = model_field.formfield()
        self.assertIsInstance(form_field, SimpleSetField)
        self.assertEqual(form_field.max_length, 4)



# -*- coding:utf-8 -*-
from threading import Thread

from django.db import connection
from django.test import TestCase
from django.utils.six.moves import queue

from django_mysql.exceptions import TimeoutError
from django_mysql.locks import Lock


class LockTests(TestCase):

    @classmethod
    def setUpClass(cls):
        super(LockTests, cls).setUpClass()

        cls.supports_lock_info = (
            connection.is_mariadb and connection.mysql_version >= (10, 0, 7)
        )
        if cls.supports_lock_info:
            with connection.cursor() as cursor:
                cursor.execute(
                    """SELECT COUNT(*) FROM INFORMATION_SCHEMA.PLUGINS
                       WHERE PLUGIN_NAME = 'metadata_lock_info'""")
                cls.lock_info_preinstalled = (cursor.fetchone()[0] > 0)
                if not cls.lock_info_preinstalled:
                    cursor.execute("INSTALL SONAME 'metadata_lock_info'")

    @classmethod
    def tearDownClass(cls):
        if cls.supports_lock_info and not cls.lock_info_preinstalled:
            with connection.cursor() as cursor:
                cursor.execute("UNINSTALL SONAME 'metadata_lock_info'")
        super(LockTests, cls).tearDownClass()

    def test_simple(self):
        mylock = Lock("mylock")
        self.assertFalse(mylock.is_held())

        with mylock:
            self.assertTrue(mylock.is_held())
            self.assertTrue(Lock("mylock").is_held())

            cursor = connection.cursor()
            cursor.execute("SELECT CONNECTION_ID();")
            own_connection_id = cursor.fetchone()[0]
            self.assertEqual(mylock.holding_connection_id(),
                             own_connection_id)

        self.assertFalse(mylock.is_held())
        self.assertFalse(Lock("mylock").is_held())

    import_time_lock = Lock('defined_at_import_time')

    def test_error_on_unneeded_exit(self):
        mylock = Lock("mylock")
        self.assertFalse(mylock.is_held())
        with self.assertRaises(ValueError) as cm:
            mylock.__exit__(None, None, None)
        self.assertIn("unheld lock", str(cm.exception))

    def test_defined_at_import_time(self):
        import_time_lock = self.import_time_lock

        self.assertFalse(import_time_lock.is_held())

        with import_time_lock:
            self.assertTrue(import_time_lock.is_held())

            cursor = connection.cursor()
            cursor.execute("SELECT CONNECTION_ID();")
            own_connection_id = cursor.fetchone()[0]
            self.assertEqual(import_time_lock.holding_connection_id(),
                             own_connection_id)

        self.assertFalse(import_time_lock.is_held())

    def test_timeout_with_threads(self):
        to_me = queue.Queue()
        to_you = queue.Queue()

        def lock_until_told():
            with Lock('threading_test'):
                to_me.put("Locked")
                to_you.get(True)

        threading_test = Lock('threading_test', 0.05)
        self.assertTrue(not threading_test.is_held())

        other_thread = Thread(target=lock_until_told)
        other_thread.start()
        try:
            item = to_me.get(True)
            self.assertEqual(item, "Locked")

            cursor = connection.cursor()
            cursor.execute("SELECT CONNECTION_ID();")
            own_connection_id = cursor.fetchone()[0]

            self.assertTrue(threading_test.is_held())
            self.assertNotEqual(threading_test.holding_connection_id(),
                                own_connection_id)

            with self.assertRaises(TimeoutError):
                with threading_test:
                        pass

            to_you.put("Stop")
        finally:
            other_thread.join()

        self.assertFalse(threading_test.is_held())
        with threading_test:
            pass

    def test_threads_concurrent_access(self):
        """
        Test that the same lock object can be used in multiple threads, allows
        the definition of a lock upfront in a module.
        """
        to_me = queue.Queue()
        to_you = queue.Queue()
        the_lock = Lock('THElock', 0.05)

        def check_it_lock_it():
            self.assertFalse(the_lock.is_held())
            with the_lock:
                to_me.put("Locked")
                to_you.get(True)

        other_thread = Thread(target=check_it_lock_it)
        other_thread.start()
        try:
            item = to_me.get(True)
            self.assertEqual(item, "Locked")

            cursor = connection.cursor()
            cursor.execute("SELECT CONNECTION_ID()")
            own_connection_id = cursor.fetchone()[0]

            self.assertTrue(the_lock.is_held())
            self.assertNotEqual(the_lock.holding_connection_id(),
                                own_connection_id)

            with self.assertRaises(TimeoutError):
                with the_lock:
                    pass

            to_you.put("Stop")
        finally:
            other_thread.join()

        with the_lock:
            pass

    def test_holding_more_than_one(self):
        conn = connection
        supports_multiple_locks = (
            (conn.is_mariadb and conn.mysql_version >= (10, 0, 2)) or
            (not conn.is_mariadb and conn.mysql_version >= (5, 7))
        )
        if not supports_multiple_locks:
            self.skipTest(
                "Only MySQL 5.7+ and MariaDB 10.0.2+ have the ability to hold "
                "more than one named lock"
            )

        lock_a = Lock("a")
        lock_b = Lock("b")
        with lock_a, lock_b:
            self.assertTrue(lock_a.is_held())

    def test_multi_connection(self):
        lock_a = Lock("a")
        lock_b = Lock("b", using='other')

        with lock_a, lock_b:
            # Different connections = can hold > 1!
            self.assertTrue(lock_a.is_held())
            self.assertTrue(lock_b.is_held())

    def test_held_with_prefix(self):
        if not self.supports_lock_info:
            self.skipTest(
                "Only MariaDB 10.0.7+ has the metadata_lock_info plugin on "
                "which held_with_prefix relies"
            )

        self.assertEqual(Lock.held_with_prefix(''), {})
        self.assertEqual(Lock.held_with_prefix('mylock'), {})

        with Lock('mylock-alpha') as lock:
            self.assertEqual(
                Lock.held_with_prefix(''),
                {'mylock-alpha': lock.holding_connection_id()}
            )
            self.assertEqual(
                Lock.held_with_prefix('mylock'),
                {'mylock-alpha': lock.holding_connection_id()}
            )
            self.assertEqual(
                Lock.held_with_prefix('mylock-beta'),
                {}
            )

        self.assertEqual(Lock.held_with_prefix(''), {})
        self.assertEqual(Lock.held_with_prefix('mylock'), {})



# -*- coding:utf-8 -*-
import json

import ddt
from django import forms
from django.core import exceptions, serializers
from django.db import models
from django.db.migrations.writer import MigrationWriter
from django.db.models import Q
from django.test import TestCase
from django.utils import six

from django_mysql.forms import SimpleSetField
from django_mysql.models import SetTextField
from django_mysql_tests.models import BigCharSetModel, BigIntSetModel


@ddt.ddt
class TestSaveLoad(TestCase):

    def test_char_easy(self):
        big_set = {six.text_type(i ** 2) for i in six.moves.range(1000)}
        s = BigCharSetModel.objects.create(field=big_set)
        self.assertSetEqual(s.field, big_set)
        s = BigCharSetModel.objects.get(id=s.id)
        self.assertSetEqual(s.field, big_set)

        letters = set('abcdefghi')
        bigger_set = big_set | letters
        s.field.update(letters)
        self.assertEqual(s.field, bigger_set)
        s.save()
        s = BigCharSetModel.objects.get(id=s.id)
        self.assertEqual(s.field, bigger_set)

    def test_char_string_direct(self):
        big_set = {six.text_type(i ** 2) for i in six.moves.range(1000)}
        big_str = ','.join(big_set)
        s = BigCharSetModel.objects.create(field=big_str)
        s = BigCharSetModel.objects.get(id=s.id)
        self.assertEqual(s.field, big_set)

    def test_is_a_set_immediately(self):
        s = BigCharSetModel()
        self.assertEqual(s.field, set())
        s.field.add("bold")
        s.field.add("brave")
        s.save()
        self.assertEqual(s.field, {"bold", "brave"})
        s = BigCharSetModel.objects.get(id=s.id)
        self.assertEqual(s.field, {"bold", "brave"})

    def test_empty(self):
        s = BigCharSetModel.objects.create()
        self.assertEqual(s.field, set())
        s = BigCharSetModel.objects.get(id=s.id)
        self.assertEqual(s.field, set())

    def test_char_cant_create_sets_with_commas(self):
        with self.assertRaises(ValueError):
            BigCharSetModel.objects.create(field={"co,mma", "contained"})

    def test_char_cant_create_sets_with_empty_string(self):
        with self.assertRaises(ValueError):
            BigCharSetModel.objects.create(field={""})

    def test_char_basic_lookup(self):
        mymodel = BigCharSetModel.objects.create()
        empty = BigCharSetModel.objects.filter(field="")

        self.assertEqual(empty.count(), 1)
        self.assertEqual(empty[0], mymodel)

        mymodel.delete()

        self.assertEqual(empty.count(), 0)

    @ddt.data('contains', 'icontains')
    def test_char_contains_lookup(self, lookup):
        lname = 'field__' + lookup
        mymodel = BigCharSetModel.objects.create(field={"mouldy", "rotten"})

        mouldy = BigCharSetModel.objects.filter(**{lname: "mouldy"})
        self.assertEqual(mouldy.count(), 1)
        self.assertEqual(mouldy[0], mymodel)

        rotten = BigCharSetModel.objects.filter(**{lname: "rotten"})
        self.assertEqual(rotten.count(), 1)
        self.assertEqual(rotten[0], mymodel)

        clean = BigCharSetModel.objects.filter(**{lname: "clean"})
        self.assertEqual(clean.count(), 0)

        with self.assertRaises(ValueError):
            list(BigCharSetModel.objects.filter(**{lname: {"a", "b"}}))

        both = BigCharSetModel.objects.filter(
            Q(**{lname: "mouldy"}) & Q(**{lname: "rotten"})
        )
        self.assertEqual(both.count(), 1)
        self.assertEqual(both[0], mymodel)

        either = BigCharSetModel.objects.filter(
            Q(**{lname: "mouldy"}) | Q(**{lname: "clean"})
        )
        self.assertEqual(either.count(), 1)

        not_clean = BigCharSetModel.objects.exclude(**{lname: "clean"})
        self.assertEqual(not_clean.count(), 1)

        not_mouldy = BigCharSetModel.objects.exclude(**{lname: "mouldy"})
        self.assertEqual(not_mouldy.count(), 0)

    def test_char_len_lookup_empty(self):
        mymodel = BigCharSetModel.objects.create(field=set())

        empty = BigCharSetModel.objects.filter(field__len=0)
        self.assertEqual(empty.count(), 1)
        self.assertEqual(empty[0], mymodel)

        one = BigCharSetModel.objects.filter(field__len=1)
        self.assertEqual(one.count(), 0)

        one_or_more = BigCharSetModel.objects.filter(field__len__gte=0)
        self.assertEqual(one_or_more.count(), 1)

    def test_char_len_lookup(self):
        mymodel = BigCharSetModel.objects.create(field={"red", "expensive"})

        empty = BigCharSetModel.objects.filter(field__len=0)
        self.assertEqual(empty.count(), 0)

        one_or_more = BigCharSetModel.objects.filter(field__len__gte=1)
        self.assertEqual(one_or_more.count(), 1)
        self.assertEqual(one_or_more[0], mymodel)

        two = BigCharSetModel.objects.filter(field__len=2)
        self.assertEqual(two.count(), 1)
        self.assertEqual(two[0], mymodel)

        three = BigCharSetModel.objects.filter(field__len=3)
        self.assertEqual(three.count(), 0)

    def test_int_easy(self):
        big_set = {i ** 2 for i in six.moves.range(1000)}
        mymodel = BigIntSetModel.objects.create(field=big_set)
        self.assertSetEqual(mymodel.field, big_set)
        mymodel = BigIntSetModel.objects.get(id=mymodel.id)
        self.assertSetEqual(mymodel.field, big_set)

    def test_int_contains_lookup(self):
        onetwo = BigIntSetModel.objects.create(field={1, 2})

        ones = BigIntSetModel.objects.filter(field__contains=1)
        self.assertEqual(ones.count(), 1)
        self.assertEqual(ones[0], onetwo)

        twos = BigIntSetModel.objects.filter(field__contains=2)
        self.assertEqual(twos.count(), 1)
        self.assertEqual(twos[0], onetwo)

        threes = BigIntSetModel.objects.filter(field__contains=3)
        self.assertEqual(threes.count(), 0)

        with self.assertRaises(ValueError):
            list(BigIntSetModel.objects.filter(field__contains={1, 2}))

        ones_and_twos = BigIntSetModel.objects.filter(
            Q(field__contains=1) & Q(field__contains=2)
        )
        self.assertEqual(ones_and_twos.count(), 1)
        self.assertEqual(ones_and_twos[0], onetwo)

        ones_and_threes = BigIntSetModel.objects.filter(
            Q(field__contains=1) & Q(field__contains=3)
        )
        self.assertEqual(ones_and_threes.count(), 0)

        ones_or_threes = BigIntSetModel.objects.filter(
            Q(field__contains=1) | Q(field__contains=3)
        )
        self.assertEqual(ones_or_threes.count(), 1)

        no_three = BigIntSetModel.objects.exclude(field__contains=3)
        self.assertEqual(no_three.count(), 1)

        no_one = BigIntSetModel.objects.exclude(field__contains=1)
        self.assertEqual(no_one.count(), 0)


class TestValidation(TestCase):

    def test_max_length(self):
        field = SetTextField(models.CharField(max_length=32), size=3)

        field.clean({'a', 'b', 'c'}, None)

        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean({'a', 'b', 'c', 'd'}, None)
        self.assertEqual(
            cm.exception.messages[0],
            'Set contains 4 items, it should contain no more than 3.'
        )


class TestCheck(TestCase):

    def test_field_checks(self):
        field = SetTextField(models.CharField())
        field.set_attributes_from_name('field')
        errors = field.check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, 'django_mysql.E001')
        self.assertIn('Base field for set has errors', errors[0].msg)
        self.assertIn('max_length', errors[0].msg)

    def test_invalid_base_fields(self):
        field = SetTextField(models.ForeignKey('django_mysql_tests.Author'))
        field.set_attributes_from_name('field')
        errors = field.check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, 'django_mysql.E002')
        self.assertIn('Base field for set must be', errors[0].msg)


class TestMigrations(TestCase):

    def test_deconstruct(self):
        field = SetTextField(models.IntegerField(), max_length=32)
        name, path, args, kwargs = field.deconstruct()
        new = SetTextField(*args, **kwargs)
        self.assertEqual(type(new.base_field), type(field.base_field))

    def test_deconstruct_with_size(self):
        field = SetTextField(models.IntegerField(), size=3, max_length=32)
        name, path, args, kwargs = field.deconstruct()
        new = SetTextField(*args, **kwargs)
        self.assertEqual(new.size, field.size)

    def test_deconstruct_args(self):
        field = SetTextField(models.CharField(max_length=5), max_length=32)
        name, path, args, kwargs = field.deconstruct()
        new = SetTextField(*args, **kwargs)
        self.assertEqual(
            new.base_field.max_length,
            field.base_field.max_length
        )

    def test_makemigrations(self):
        field = SetTextField(models.CharField(max_length=5))
        statement, imports = MigrationWriter.serialize(field)

        self.assertEqual(
            statement,
            "django_mysql.models.SetTextField("
            "models.CharField(max_length=5), "
            "size=None"
            ")"
        )

    def test_makemigrations_with_size(self):
        field = SetTextField(models.CharField(max_length=5), size=5)
        statement, imports = MigrationWriter.serialize(field)

        self.assertEqual(
            statement,
            "django_mysql.models.SetTextField("
            "models.CharField(max_length=5), "
            "size=5"
            ")"
        )


class TestSerialization(TestCase):

    def test_dumping(self):
        big_set = {six.text_type(i ** 2) for i in six.moves.range(1000)}
        instance = BigCharSetModel(field=big_set)
        data = json.loads(serializers.serialize('json', [instance]))[0]
        field = data['fields']['field']
        self.assertEqual(sorted(field.split(',')), sorted(big_set))

    def test_loading(self):
        test_data = '''
            [{"fields": {"field": "big,leather,comfy"},
             "model": "django_mysql_tests.BigCharSetModel", "pk": null}]
        '''
        objs = list(serializers.deserialize('json', test_data))
        instance = objs[0].object
        self.assertEqual(instance.field, {"big", "leather", "comfy"})


class TestDescription(TestCase):

    def test_char(self):
        field = SetTextField(models.CharField(max_length=5), max_length=32)
        self.assertEqual(
            field.description,
            "Set of String (up to %(max_length)s)"
        )

    def test_int(self):
        field = SetTextField(models.IntegerField(), max_length=32)
        self.assertEqual(field.description, "Set of Integer")


class TestFormField(TestCase):

    def test_model_field_formfield(self):
        model_field = SetTextField(models.CharField(max_length=27))
        form_field = model_field.formfield()
        self.assertIsInstance(form_field, SimpleSetField)
        self.assertIsInstance(form_field.base_field, forms.CharField)
        self.assertEqual(form_field.base_field.max_length, 27)

    def test_model_field_formfield_size(self):
        model_field = SetTextField(models.IntegerField(), size=400)
        form_field = model_field.formfield()
        self.assertIsInstance(form_field, SimpleSetField)
        self.assertEqual(form_field.max_length, 400)



# -*- coding:utf-8 -*-
from django.test import TestCase

from django_mysql_tests.models import Author


class CaseExactTests(TestCase):

    def test_charfield(self):
        dickie = Author.objects.create(name="Dickens")

        authors = Author.objects.filter(name__case_exact="dickens")
        self.assertEqual(list(authors), [])

        authors = Author.objects.filter(name__case_exact="Dickens")
        self.assertEqual(list(authors), [dickie])

        authors = Author.objects.filter(name__case_exact="DICKENS")
        self.assertEqual(list(authors), [])

    def test_charfield_lowercase(self):
        dickie = Author.objects.create(name="dickens")

        authors = Author.objects.filter(name__case_exact="dickens")
        self.assertEqual(list(authors), [dickie])

        authors = Author.objects.filter(name__case_exact="Dickens")
        self.assertEqual(list(authors), [])

    def test_textfield(self):
        dickie = Author.objects.create(name="Dickens", bio="Aged 10, bald.")

        authors = Author.objects.filter(bio__case_exact="Aged 10, bald.")
        self.assertEqual(list(authors), [dickie])

        authors = Author.objects.filter(bio__case_exact="Aged 10, BALD.")
        self.assertEqual(list(authors), [])


class SoundexTests(TestCase):

    def test_sounds_like_lookup(self):
        principles = ["principle", "principal", "princpl"]
        created = {Author.objects.create(name=name) for name in principles}

        for name in principles:
            sounding = Author.objects.filter(name__sounds_like=name)
            self.assertEqual(set(sounding), created)

        sounding = Author.objects.filter(name__sounds_like='')
        self.assertEqual(set(sounding), set())

        sounding = Author.objects.filter(name__sounds_like='nothing')
        self.assertEqual(set(sounding), set())

    def test_soundex_strings(self):
        author = Author.objects.create(name='Robert')
        self.assertEqual(Author.objects.get(name__soundex='R163'), author)



# -*- coding:utf-8 -*-
import mock
from unittest import skipUnless

from django.db.models.query import QuerySet
from django.template import Context, Template
from django.test import TransactionTestCase

from django_mysql.models import ApproximateInt, SmartIterator
from django_mysql.utils import have_program
from django_mysql_tests.models import Author, NameAuthor, VanillaAuthor
from django_mysql_tests.utils import captured_stdout


class ApproximateCountTests(TransactionTestCase):

    def setUp(self):
        super(ApproximateCountTests, self).setUp()
        Author.objects.bulk_create([Author() for i in range(10)])

    def test_approx_count(self):
        # Theoretically this varies 30-50% of the table size
        # For a fresh table with 10 items we seem to always get back the actual
        # count, but to be sure we'll just assert it's within 55%
        approx_count = Author.objects.approx_count(min_size=1)
        self.assertGreaterEqual(approx_count, 4)
        self.assertLessEqual(approx_count, 16)

    def test_activation_deactivation(self):
        qs = Author.objects.all()
        self.assertFalse(qs._count_tries_approx)

        qs2 = qs.count_tries_approx(min_size=2)
        self.assertNotEqual(qs, qs2)
        self.assertTrue(qs2._count_tries_approx)
        count = qs2.count()
        self.assertTrue(isinstance(count, ApproximateInt))

        qs3 = qs2.count_tries_approx(False)
        self.assertNotEqual(qs2, qs3)
        self.assertFalse(qs3._count_tries_approx)

    def test_activation_but_fallback(self):
        qs = Author.objects.exclude(name='TEST').count_tries_approx()
        count = qs.count()
        self.assertEqual(count, 10)
        self.assertFalse(isinstance(count, ApproximateInt))

    def test_activation_but_fallback_due_to_min_size(self):
        qs = Author.objects.count_tries_approx()
        count = qs.count()
        self.assertEqual(count, 10)
        self.assertFalse(isinstance(count, ApproximateInt))

    def test_output_in_templates(self):
        approx_count = Author.objects.approx_count(min_size=1)
        text = Template('{{ var }}').render(Context({'var': approx_count}))
        self.assertTrue(text.startswith('Approximately '))

        approx_count2 = Author.objects.approx_count(
            min_size=1,
            return_approx_int=False
        )
        text = Template('{{ var }}').render(Context({'var': approx_count2}))
        self.assertFalse(text.startswith('Approximately '))

    def test_fallback_with_filters(self):
        filtered = Author.objects.filter(name='')
        self.assertEqual(filtered.approx_count(fall_back=True), 10)
        with self.assertRaises(ValueError):
            filtered.approx_count(fall_back=False)

    def test_fallback_with_slice(self):
        self.assertEqual(Author.objects.all()[:100].approx_count(), 10)
        with self.assertRaises(ValueError):
            Author.objects.all()[:100].approx_count(fall_back=False)

    def test_fallback_with_distinct(self):
        self.assertEqual(Author.objects.distinct().approx_count(), 10)
        with self.assertRaises(ValueError):
            Author.objects.distinct().approx_count(fall_back=False)


class SmartIteratorTests(TransactionTestCase):

    def setUp(self):
        super(SmartIteratorTests, self).setUp()
        Author.objects.bulk_create([Author() for i in range(10)])

    def test_bad_querysets(self):
        with self.assertRaises(ValueError) as cm:
            Author.objects.all().order_by('name').iter_smart_chunks()
        self.assertIn("ordering", str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            Author.objects.all()[:5].iter_smart_chunks()
        self.assertIn("sliced QuerySet", str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            NameAuthor.objects.all().iter_smart_chunks()
        self.assertIn("non-integer primary key", str(cm.exception))

    def test_chunks(self):
        seen = []
        for authors in Author.objects.iter_smart_chunks():
            seen.extend(author.id for author in authors)

        all_ids = list(Author.objects.order_by('id')
                                     .values_list('id', flat=True))
        self.assertEqual(seen, all_ids)

    def test_objects(self):
        seen = [author.id for author in Author.objects.iter_smart()]
        all_ids = list(Author.objects.order_by('id')
                                     .values_list('id', flat=True))
        self.assertEqual(seen, all_ids)

    def test_objects_non_atomic(self):
        seen = [author.id for author in
                Author.objects.iter_smart(atomically=False)]
        all_ids = list(Author.objects.order_by('id')
                                     .values_list('id', flat=True))
        self.assertEqual(seen, all_ids)

    def test_objects_pk_range_all(self):
        seen = [author.id for author in
                Author.objects.iter_smart(pk_range='all')]
        all_ids = list(Author.objects.order_by('id')
                                     .values_list('id', flat=True))
        self.assertEqual(seen, all_ids)

    def test_objects_pk_range_tuple(self):
        seen = [author.id for author in
                Author.objects.iter_smart(pk_range=(0, 0))]
        self.assertEqual(seen, [])

        min_id = Author.objects.earliest('id').id
        max_id = Author.objects.order_by('id')[5].id

        seen = [author.id for author in
                Author.objects.iter_smart(pk_range=(min_id, max_id))]
        cut_ids = list(Author.objects.order_by('id')
                                     .filter(id__gte=min_id, id__lte=max_id)
                                     .values_list('id', flat=True))
        self.assertEqual(seen, cut_ids)

    def test_objects_pk_range_bad(self):
        with self.assertRaises(ValueError) as cm:
            list(Author.objects.iter_smart(pk_range="My Bad Value"))
        self.assertIn("Unrecognized value for pk_range", str(cm.exception))

    def test_pk_range_race_condition(self):
        getitem = QuerySet.__getitem__

        def fail_second_slice(*args, **kwargs):
            # Simulate race condition by deleting all objects between first
            # call (min_qs[0]) and second call (max_qs[0]) to
            # QuerySet.__getitem__
            fail_second_slice.calls += 1
            if fail_second_slice.calls == 2:
                Author.objects.all().delete()
            return getitem(*args, **kwargs)

        fail_second_slice.calls = 0

        path = 'django.db.models.query.QuerySet.__getitem__'

        with mock.patch(path, fail_second_slice):
            seen = [author.id for author in Author.objects.iter_smart()]
        self.assertEqual(seen, [])

    def test_objects_max_size(self):
        seen = [author.id for author in
                Author.objects.iter_smart(chunk_max=1)]
        all_ids = list(Author.objects.order_by('id')
                                     .values_list('id', flat=True))
        self.assertEqual(seen, all_ids)

    def test_no_matching_objects(self):
        seen = [author.id for author in
                Author.objects.filter(name="Waaa").iter_smart()]
        self.assertEqual(seen, [])

    def test_no_objects(self):
        Author.objects.all().delete()
        seen = [author.id for author in Author.objects.iter_smart()]
        self.assertEqual(seen, [])

    def test_pk_hole(self):
        first = Author.objects.earliest('id')
        last = Author.objects.latest('id')
        Author.objects.filter(id__gt=first.id, id__lt=last.id).delete()
        seen = [author.id for author in Author.objects.iter_smart()]
        self.assertEqual(seen, [first.id, last.id])

    def test_reporting(self):
        with captured_stdout() as output:
            qs = Author.objects.all()
            for authors in qs.iter_smart_chunks(report_progress=True):
                list(authors)  # fetch them

        lines = output.getvalue().split('\n')

        reports = lines[0].split('\r')
        for report in reports:
            self.assertRegexpMatches(
                report,
                r"AuthorSmartChunkedIterator processed \d+/10 objects "
                r"\(\d+\.\d+%\) in \d+ chunks(; highest pk so far \d+)?"
            )

        self.assertEqual(lines[1], 'Finished!')

    def test_reporting_with_total(self):
        with captured_stdout() as output:
            qs = Author.objects.all()
            for authors in qs.iter_smart_chunks(report_progress=True, total=4):
                list(authors)  # fetch them

        lines = output.getvalue().split('\n')

        reports = lines[0].split('\r')
        for report in reports:
            self.assertRegexpMatches(
                report,
                r"AuthorSmartChunkedIterator processed \d+/4 objects "
                r"\(\d+\.\d+%\) in \d+ chunks(; highest pk so far \d+)?"
            )

        self.assertEqual(lines[1], 'Finished!')

    def test_reporting_on_uncounted_qs(self):
        Author.objects.create(name="pants")

        with captured_stdout() as output:
            qs = Author.objects.filter(name="pants")
            for authors in qs.iter_smart_chunks(report_progress=True):
                authors.delete()

        lines = output.getvalue().split('\n')

        reports = lines[0].split('\r')
        for report in reports:
            self.assertRegexpMatches(
                report,
                # We should have ??? since the deletion means the objects
                # aren't fetched into python
                r"AuthorSmartChunkedIterator processed (0|\?\?\?)/1 objects "
                r"\(\d+\.\d+%\) in \d+ chunks(; highest pk so far \d+)?"
            )

        self.assertEqual(lines[1], 'Finished!')

    def test_running_on_non_mysql_model(self):
        VanillaAuthor.objects.create(name="Alpha")
        VanillaAuthor.objects.create(name="pants")
        VanillaAuthor.objects.create(name="Beta")
        VanillaAuthor.objects.create(name="pants")

        bad_authors = VanillaAuthor.objects.filter(name="pants")

        self.assertEqual(bad_authors.count(), 2)

        with captured_stdout():
            for author in SmartIterator(bad_authors, report_progress=True):
                author.delete()

        self.assertEqual(bad_authors.count(), 0)


@skipUnless(have_program('pt-visual-explain'),
            "pt-visual-explain must be installed")
class VisualExplainTests(TransactionTestCase):

    def test_basic(self):
        with captured_stdout() as capture:
            Author.objects.all().pt_visual_explain()
        output = capture.getvalue()
        self.assertGreater(output, "")
        # Can't be too strict about the output since different database and pt-
        # visual-explain versions give different output
        self.assertIn("django_mysql_tests_author", output)
        self.assertIn("rows", output)
        self.assertIn("Table", output)

    def test_basic_no_display(self):
        output = Author.objects.all().pt_visual_explain(display=False)
        self.assertGreater(output, "")
        self.assertIn("django_mysql_tests_author", output)
        self.assertIn("rows", output)
        self.assertIn("Table", output)

    def test_subquery(self):
        subq = Author.objects.all().values_list('id', flat=True)
        output = Author.objects.filter(id__in=subq) \
                               .pt_visual_explain(display=False)
        self.assertGreater(output, "")
        self.assertIn("possible_keys", output)
        self.assertIn("django_mysql_tests_author", output)
        self.assertIn("rows", output)
        self.assertIn("Table", output)



# -*- coding:utf-8 -*-
from __future__ import unicode_literals

from django.utils.functional import cached_property


@cached_property
def is_mariadb(self):
    """
    Method to be monkey-patched onto connection, to allow easy checks for if
    the current database connection is MariaDB
    """
    with self.temporary_connection():
        server_info = self.connection.get_server_info()
        return 'MariaDB' in server_info
